@@ -5,11 +5,13 @@
 #' @param formula A formula.
 #' Its structure is similar to \code{\link[lme4]{lmer}} function
 #' in **lme4** package.
-#' @param data A DataFrame.
-#' @param local.fixed A character vector.
-#' It contains names of local fixed effects.
-#' @param coords A 2-column matrix.
-#' It consists of coordinates for each group.
+#' Models can be specified with the following form:
+#' ```r
+#' response ~ L(local.fixed) + global.fixed + (random | group)
+#' ```
+#' For more information, please see the `formula` subsection in details.
+#' @param data The data.
+#' @param \dots Further arguments for the specified type of `data`.
 #' @param bw A numeric value. It is the value of bandwidth or `"CV"`.
 #' In this stage this function only support adaptive bandwidth.
 #' And its unit must be the number of nearest neighbours.
@@ -56,17 +58,100 @@
 #'  \item{\code{frame.parsed}}{Variables extracted from the data.}
 #'  \item{\code{groups}}{Unique group labels extracted from the data.}
 #' }
+#' 
+#' @details  
+#' ## Effect Specification in Formula
+#' In the HGWR model, there are three types of effects specified by the
+#' `formula` argument:
+#' \describe{
+#'  \item{Local fixed effects}{Effects wrapped by functional symbol `L`.}
+#'  \item{Random effects}{Effects specified outside the functional symbol `L` but to the left of symbol `|`.}
+#'  \item{Global fixed effects}{Other effects}
+#' }
+#' For example, the following formula in the example of this function below is written as
+#' ```r
+#' y ~ L(g1 + g2) + x1 + (z1 | group)
+#' ```
+#' where `g1` and `g2` are local fixed effects,
+#' `x1` is the global fixed effects,
+#' and `z1` is the random effects grouped by the group indicator `group`.
+#' Note that random effects can only be specified once!
 #'
 #' @examples
 #' data(multisampling)
-#' hgwr(formula = y ~ g1 + g2 + x1 + (z1 | group),
+#' hgwr(formula = y ~ L(g1 + g2) + x1 + (z1 | group),
 #'      data = multisampling$data,
-#'      local.fixed = c("g1", "g2"),
 #'      coords = multisampling$coords,
 #'      bw = 10)
-#'
+#' 
+#' @importFrom stats aggregate
+#' 
+#' @export 
 hgwr <- function(
-    formula, data, local.fixed, coords, bw = "CV",
+    formula, data, ..., bw = "CV",
+    kernel = c("gaussian", "bisquared"),
+    alpha = 0.01, eps_iter = 1e-6, eps_gradient = 1e-6,
+    max_iters = 1e6, max_retries = 1e6,
+    ml_type = c("D_Only", "D_Beta"), verbose = 0
+) {
+    UseMethod("hgwr", data)
+}
+
+#' @rdname hgwr
+#' @method hgwr sf
+#' 
+#' @importFrom sf st_centroid st_coordinates
+#' @importFrom stats aggregate
+#' 
+#' @export 
+hgwr.sf <- function(
+    formula, data, ..., bw = "CV",
+    kernel = c("gaussian", "bisquared"),
+    alpha = 0.01, eps_iter = 1e-6, eps_gradient = 1e-6,
+    max_iters = 1e6, max_retries = 1e6,
+    ml_type = c("D_Only", "D_Beta"), verbose = 0
+) {
+    ### Generate group-level coordinates by taking means
+    data_coords <- sf::st_coordinates(sf::st_centroid(data))
+    data <- sf::st_drop_geometry(data)
+    group <- data[[parse.formula(formula)$group]]
+    group_unique <- unique(group)
+    group_index <- as.vector(match(group, group_unique))
+    group_coords <- aggregate(data_coords, by = list(group_index), FUN = mean)[,-1]
+    mc0 <- mc <- match.call(expand.dots = TRUE)
+    mc[[1]] <- as.name("hgwr_fit")
+    mc[["data"]] <- data
+    mc[["coords"]] <- group_coords
+    mev <- eval.parent(mc)
+    mev$call <- mc0
+    mev
+}
+
+#' @rdname hgwr
+#' @method hgwr data.frame
+#' 
+#' @param coords A 2-column matrix.
+#' It consists of coordinates for each group.
+#' 
+#' @export 
+hgwr.data.frame <- function(
+    formula, data, ..., coords, bw = "CV",
+    kernel = c("gaussian", "bisquared"),
+    alpha = 0.01, eps_iter = 1e-6, eps_gradient = 1e-6,
+    max_iters = 1e6, max_retries = 1e6,
+    ml_type = c("D_Only", "D_Beta"), verbose = 0
+) {
+    mc0 <- mc <- match.call(expand.dots = TRUE)
+    mc[[1]] <- as.name("hgwr_fit")
+    mev <- eval.parent(mc)
+    mev$call <- mc0
+    mev
+}
+
+#' @describeIn hgwr Fit a HGWR model
+#' @export 
+hgwr_fit <- function(
+    formula, data, coords, bw = "CV",
     kernel = c("gaussian", "bisquared"),
     alpha = 0.01, eps_iter = 1e-6, eps_gradient = 1e-6,
     max_iters = 1e6, max_retries = 1e6,
@@ -74,6 +159,11 @@ hgwr <- function(
 ) {
     ### Extract variables
     kernel <- match.arg(kernel)
+    kernel_index <- switch(
+        kernel,
+        "gaussian" = 0L,
+        "bisquared" = 1L
+    )
     ml_type <- switch(
         match.arg(ml_type),
         "D_Only" = 0L,
@@ -81,19 +171,21 @@ hgwr <- function(
     )
     model_desc <- parse.formula(formula)
     y <- as.vector(data[[model_desc$response]])
-    group <- as.vector(as.integer(data[[model_desc$group]]))
+    group <- data[[model_desc$group]]
     group_unique <- unique(group)
-    group_index <- match(group, group_unique)
-    z <- as.matrix(cbind(1, data[model_desc$random.effects]))
-    fe <- model_desc$fixed.effects
-    lfe <- fe[fe %in% local.fixed]
-    gfe <- fe[!(fe %in% local.fixed)]
-    x <- as.matrix(cbind(1, data[gfe]))
-    g <- as.matrix(cbind(1, aggregate(data[lfe], list(group), mean)[,-1]))
+    group_index <- as.vector(match(group, group_unique))
+    z <- as.matrix(make.dummy(data[model_desc$random.effects]))
+    if (model_desc$intercept$random) z <- cbind(1, z)
+    gfe <- model_desc$fixed.effects
+    lfe <- model_desc$local.fixed.effects
+    x <- as.matrix(make.dummy(data[gfe]))
+    if (model_desc$intercept$fixed) x <- cbind(1, x)
+    g <- as.matrix(aggregate(make.dummy(data[lfe]), list(group), mean)[,-1])
+    if (model_desc$intercept$local) g <- cbind(1, g)
 
     ### Get bandwidth value
     if (is.character(bw) && bw == "CV") {
-        bw_value <- 0.0
+        bw_value <- NA_real_
         optim_bw <- TRUE
     } else if (is.numeric(bw) || is.integer(bw)) {
         bw_value <- bw
@@ -104,28 +196,30 @@ hgwr <- function(
     }
 
     ### Call C
-    hgwr_result <- .hgwr_bml(
-        g, x, z, y, as.matrix(coords), group_index, bw_value, kernel,
-        alpha, eps_iter, eps_gradient,
-        as.integer(max_iters), as.integer(max_retries),
-        as.integer(ml_type), as.integer(verbose)
-    )
-    gamma <- hgwr_result$gamma
-    beta <- t(hgwr_result$beta)
-    mu <- hgwr_result$mu
-    D <- hgwr_result$D
-    sigma <- hgwr_result$sigma
+    hgwr_result <- tryCatch({
+        hgwr_bfml(
+            g, x, z, y, as.matrix(coords), group_index, bw_value, kernel_index,
+            alpha, eps_iter, eps_gradient,
+            as.integer(max_iters), as.integer(max_retries),
+            as.integer(ml_type), as.integer(verbose)
+        )
+    }, error = function(e) {
+        stop("Error occurred when estimating HGWR parameters.")
+    })
     if (optim_bw)
         bw_value <- hgwr_result$bw
 
     ### Prepare Return Result
     result <- list(
-        gamma = gamma,
-        beta = beta,
-        mu = mu,
-        D = D,
-        sigma = sigma,
+        gamma = hgwr_result$gamma,
+        beta = hgwr_result$beta,
+        mu = hgwr_result$mu,
+        D = hgwr_result$D,
+        sigma = hgwr_result$sigma,
         bw = bw_value,
+        logLik = hgwr_result$logLik,
+        trS = hgwr_result$trS,
+        var_beta = hgwr_result$var_beta,
         effects = list(
             global.fixed = gfe,
             local.fixed = lfe,
@@ -133,16 +227,17 @@ hgwr <- function(
             group = model_desc$group,
             response = model_desc$response
         ),
-        call = match.call(),
+        intercept = model_desc$intercept,
         frame = data,
         frame.parsed = list(
             y = y,
             x = x,
             g = g,
             z = z,
-            group = group
+            group = group_index
         ),
-        groups = group_unique
+        groups = group_unique,
+        coords = coords
     )
     class(result) <- "hgwrm"
     result
@@ -156,34 +251,55 @@ hgwr <- function(
 #' @return A \code{DataFrame} object consists of all estimated coefficients.
 #'
 #' @seealso [hgwr()], [summary.hgwrm()], [fitted.hgwrm()] and [residuals.hgwrm()].
+#' 
+#' @export
 coef.hgwrm <- function(object, ...) {
     if (!inherits(object, "hgwrm")) {
         stop("It's not a hgwrm object.")
     }
     gamma <- object$gamma
-    beta <- matrix(object$beta, nrow = length(object$groups), ncol = ncol(object$beta), byrow = T)
+    beta <- matrix(object$beta, nrow = length(object$groups), ncol = length(object$beta), byrow = T)
     mu <- object$mu
-    intercept <- gamma[,1] + mu[,1] + beta[,1]
+    intercept <- matrix(0, length(object$groups), 1)
+    if (object$intercept$local) {
+        intercept <- intercept + gamma[,1]
+        gamma <- gamma[,-1]
+    }
+    if (object$intercept$fixed) {
+        intercept <- intercept + beta[,1]
+        beta <- beta[,-1]
+    }
+    if (object$intercept$random) {
+        intercept <- intercept + mu[,1]
+        mu <- mu[,-1]
+    }
     effects <- object$effects
-    coef <- as.data.frame(cbind(intercept, gamma[,-1], beta[,-1], mu[,-1], object$groups))
-    colnames(coef) <- c("Intercept", effects$local.fixed, effects$global.fixed, effects$random, effects$group)
+    coef <- as.data.frame(cbind(intercept, gamma, beta, mu))
+    coef_names <- c(effects$local.fixed, effects$global.fixed, effects$random)
+    if (any(unlist(object$intercept))) {
+        colnames(coef) <- c("Intercept", coef_names)
+    } else {
+        colnames(coef) <- coef_names
+    }
     coef
 }
 
-#' Get fitted reponse.
+#' Get fitted response.
 #'
 #' @inheritParams coef.hgwrm
 #'
 #' @return A vector consists of fitted response values.
 #'
 #' @seealso [hgwr()], [summary.hgwrm()], [coef.hgwrm()] and [residuals.hgwrm()].
+#' 
+#' @export 
 fitted.hgwrm <- function(object, ...) {
     if (!inherits(object, "hgwrm")) {
         stop("It's not a hgwrm object.")
     }
     xf <- object$frame.parsed
     rowSums(xf$g * object$gamma)[xf$group] +
-        as.vector(xf$x %*% t(object$beta)) +
+        as.vector(xf$x %*% object$beta) +
         rowSums(xf$z * object$mu[xf$group,])
 }
 
@@ -194,6 +310,9 @@ fitted.hgwrm <- function(object, ...) {
 #' @return A vector consists of residuals.
 #'
 #' @seealso [hgwr()], [summary.hgwrm()], [coef.hgwrm()] and [fitted.hgwrm()].
+#' 
+#' @export 
+#' 
 residuals.hgwrm <- function(object, ...) {
     if (!inherits(object, "hgwrm")) {
         stop("It's not a hgwrm object.")
@@ -205,6 +324,11 @@ residuals.hgwrm <- function(object, ...) {
 #'
 #' @param object An `hgwrm` object returned from [hgwr()].
 #' @param \dots Other arguments passed from other functions.
+#' @param test_hetero Logical/list value.
+#' Whether to test the spatial heterogeneity of local fixed effects.
+#' If it is set to `FALSE`, the test will not be executed.
+#' If it is set to `TRUE`, the test will be executed with default parameters (see details below).
+#' It accepts a list to enable the test with specified parameters.
 #'
 #' @return A list containing summary informations of this `hgwrm` object
 #' with the following fields.
@@ -214,10 +338,18 @@ residuals.hgwrm <- function(object, ...) {
 #'  \item{\code{random.corr}}{The correlation matrix of random effects.}
 #'  \item{\code{residuals}}{The residual vector.}
 #' }
-#'
+#' 
+#' @details The parameters used to perform test of spatial heterogeneity are
+#' \describe{
+#'  \item{\code{bw}}{Bandwidth (unit: number of nearest neighbours) used to make spatial kernel density estimation. Default: `10`.}
+#'  \item{\code{poly}}{The number of polynomial terms used in the local polynomial estimation. Default: `2`.}
+#'  \item{\code{resample}}{Total resampling times. Default: `5000`.}
+#' }
+#' 
+#' @importFrom stats AIC logLik pt sd
 #' @seealso [hgwr()].
-#'
-summary.hgwrm <- function(object, ...) {
+#' @export 
+summary.hgwrm <- function(object, ..., test_hetero = FALSE) {
     if (!inherits(object, "hgwrm")) {
         stop("It's not a hgwrm object.")
     }
@@ -225,15 +357,61 @@ summary.hgwrm <- function(object, ...) {
     res <- as.list(object)
 
     ### Diagnostics
+    #### R-squared
     y <- object$frame[[object$effects$response]]
     tss <- sum((y - mean(y))^2)
     x_residuals <- residuals.hgwrm(object)
     rss <- sum(x_residuals^2)
     rsquared <- 1 - rss / tss
+    #### AIC
+    logLik_object <- stats::logLik(object)
+    aic <- stats::AIC(logLik_object)
+    #### Record results
     res$diagnostic <- list(
-        rsquared = rsquared
+        rsquared = rsquared,
+        logLik = object$logLik,
+        AIC = aic
     )
 
+    ### Significance test
+    significance <- list()
+    #### Beta
+    enp <- attr(logLik_object, "df")
+    edf <- length(y) - enp
+    se_beta <- sqrt(object$var_beta)
+    t_beta <- abs(object$beta) / se_beta
+    p_beta <- (1 - stats::pt(t_beta, edf)) * 2
+    significance$beta <- data.frame(
+        est = object$beta,
+        sd = se_beta,
+        tv = t_beta,
+        pv = p_beta
+    )
+    #### Gamma
+    if (test_hetero == TRUE || is.list(test_hetero)) {
+        bw <- 10L
+        resample <- 5000L
+        poly <- 2L
+        if (is.list(test_hetero)) {
+            bw <- ifelse("bw" %in% names(test_hetero), test_hetero$bw, bw)
+            resample <- ifelse("resample" %in% names(test_hetero), test_hetero$resample, resample)
+            poly <- ifelse("poly" %in% names(test_hetero), test_hetero$poly, poly)
+        }
+        mean_gamma <- colMeans(object$gamma)
+        sd_gamma <- apply(object$gamma, 2, stats::sd)
+        t_gamma <- spatial_hetero_perm(object$gamma, as.matrix(object$coords), poly = poly, resample = resample, bw = bw)
+        pv <- sapply(seq_along(t_gamma$t0), function(i) {
+            with(t_gamma, mean(t[,i] > t0[i]))
+        })
+        significance$gamma <- data.frame(
+            mean = mean_gamma,
+            sd = sd_gamma,
+            pv = pv
+        )
+    }
+    #### Save results
+    res$significance <- significance
+
     ### Random effects
     random_corr_cov <- object$sigma * object$sigma * object$D
     random_stddev <- sqrt(diag(random_corr_cov))
@@ -250,112 +428,6 @@ summary.hgwrm <- function(object, ...) {
     res
 }
 
-#' Print a character matrix as a table.
-#'
-#' @param x A character matrix.
-#' @param col.sep Column seperator. Default to `""`.
-#' @param header.sep Header seperator. Default to `"-"`.
-#' @param row.begin Character at the beginning of each row.
-#' Default to `col.sep`.
-#' @param row.end Character at the ending of each row.
-#' Default to `col.sep`.
-#' @param table.style Name of pre-defined style.
-#' Possible values are `"plain"`, `"md"` or `"latex"`. Default to `"plain"`.
-#' @param \dots Additional style control arguments.
-#'
-#' @return No return.
-#'
-#' @details
-#' When `table.style` is specified, `col.sep`, `header.sep`, `row.begin`
-#' and `row.end` would not take effects.
-#' Because this function will automatically set their values.
-#' For each possible value of `table.style`, its corresponding style settings
-#' are shown in the following table.
-#' \tabular{llll}{
-#'                   \tab \strong{\code{plain}} \tab \strong{\code{md}} \tab \strong{\code{latex}} \cr
-#' \code{col.sep}    \tab \code{""}             \tab \code{"|"}         \tab \code{"&"}            \cr
-#' \code{header.sep} \tab \code{""}             \tab \code{"-"}         \tab \code{""}             \cr
-#' \code{row.begin}  \tab \code{""}             \tab \code{"|"}         \tab \code{""}             \cr
-#' \code{row.end}    \tab \code{""}             \tab \code{"|"}         \tab \code{"\\\\"}
-#' }
-#'
-#' In this function, characters are right padded by spaces.
-#'
-#' @seealso [print.hgwrm()], [summary.hgwrm()].
-print.table.md <- function(x, col.sep = "", header.sep = "",
-                           row.begin = "", row.end = "",
-                           table.style = c("plain", "md", "latex"), ...) {
-    if (!missing(table.style)) {
-        table.style <- match.arg(table.style)
-        if (table.style == "md") {
-            col.sep <- "|"
-            header.sep <- "-"
-            row.begin <- "|"
-            row.end <- "|"
-        } else if (table.style == "latex") {
-            col.sep <- "&"
-            header.sep <- ""
-            row.begin <- ""
-            row.end <- "\\\\"
-        } else if (table.style == "plain") {
-            col.sep <- ""
-            header.sep <- ""
-            row.begin <- ""
-            row.end <- ""
-        } else {
-           stop("Unknown table.style.")
-        }
-    }
-    if (nchar(header.sep) > 1) {
-       stop("Currently only 1 character header.sep is supported.")
-    }
-    ### Print table
-    x.length <- apply(x, c(1, 2), nchar)
-    x.length.max <- apply(x.length, 2, max)
-    x.fmt <- sprintf("%%%ds", x.length.max)
-    for(c in 1:ncol(x)) {
-        if(x.length.max[c] > 0)
-            cat(ifelse(c == 1, row.begin, col.sep),
-                sprintf(x.fmt[c], x[1, c]), "")
-    }
-    cat(paste0(row.end, "\n"))
-    if (nchar(header.sep) > 0) {
-        for(c in 1:ncol(x)) {
-            if(x.length.max[c] > 0) {
-                header.sep.full <- paste(rep("-", x.length.max[c]),
-                                         collapse = "")
-                cat(ifelse(c == 1, row.begin, col.sep),
-                    sprintf(header.sep.full), "")
-            }
-        }
-        cat(paste0(row.end, "\n"))
-    }
-    for (r in 2:nrow(x)) {
-        for (c in 1:ncol(x)) {
-            if(x.length.max[c] > 0)
-                cat(ifelse(c == 1, row.begin, col.sep),
-                    sprintf(x.fmt[c], x[r, c]), "")
-        }
-        cat(paste0(row.end, "\n"))
-    }
-}
-
-#' Convert a numeric matrix to character matrix according to a format string.
-#'
-#' @param m A numeric matrix.
-#' @param fmt Format string. Passing to [base::sprintf()].
-#'
-#' @seealso [base::sprintf()], [print.hgwrm()], [print.summary.hgwrm()].
-matrix2char <- function(m, fmt = "%.6f") {
-    mc <- NULL
-    if ("array" %in% class(m)) {
-        mc <- apply(m, seq(length(dim(m))), function(x) { sprintf(fmt, x) })
-    } else {
-        mc <- sprintf(fmt, m)
-    }
-    mc
-}
-
 #' Print description of a `hgwrm` object.
 #'
 #' @param x An `hgwrm` object returned by [hgwr()].
@@ -366,15 +438,18 @@ matrix2char <- function(m, fmt = "%.6f") {
 #'
 #' @examples
 #' data(multisampling)
-#' model <- hgwr(formula = y ~ g1 + g2 + x1 + (z1 | group),
+#' model <- hgwr(formula = y ~ L(g1 + g2) + x1 + (z1 | group),
 #'               data = multisampling$data,
-#'               local.fixed = c("g1", "g2"),
 #'               coords = multisampling$coords,
 #'               bw = 10)
 #' print(model)
 #' print(model, table.style = "md")
 #'
 #' @seealso [summary.hgwrm()], [print.table.md()].
+#' 
+#' @importFrom stats fivenum
+#' 
+#' @export 
 #'
 print.hgwrm <- function(x, decimal.fmt = "%.6f", ...) {
     if (!inherits(x, "hgwrm")) {
@@ -389,11 +464,15 @@ print.hgwrm <- function(x, decimal.fmt = "%.6f", ...) {
     cat("   Data:", deparse(x$call[[3]]), fill = T)
     cat("\n")
     effects <- x$effects
+    intercept <- x$intercept
+    if (intercept$fixed) effects$global.fixed <- c("Intercept", effects$global.fixed)
+    if (intercept$local) effects$local.fixed <- c("Intercept", effects$local.fixed)
+    if (intercept$random) effects$random <- c("Intercept", effects$random)
     cat("Global Fixed Effects", fill = T)
     cat("-------------------", fill = T)
     beta_str <- rbind(
-        c("Intercept", effects$global.fixed),
-        matrix2char(x$beta)
+        effects$global.fixed,
+        matrix2char(t(x$beta))
     )
     print.table.md(beta_str, ...)
     cat("\n")
@@ -403,7 +482,7 @@ print.hgwrm <- function(x, decimal.fmt = "%.6f", ...) {
     gamma_fivenum <- t(apply(x$gamma, 2, fivenum))
     gamma_str <- rbind(
         c("Coefficient", "Min", "1st Quartile", "Median", "3rd Quartile", "Max"),
-        cbind(c("Intercept", effects$local.fixed), matrix2char(gamma_fivenum))
+        cbind(effects$local.fixed, matrix2char(gamma_fivenum))
     )
     print.table.md(gamma_str, ...)
     cat("\n")
@@ -417,7 +496,7 @@ print.hgwrm <- function(x, decimal.fmt = "%.6f", ...) {
     random_corr_str <- rbind("", random_corr_str)
     random_corr_str[1, 1] <- "Corr"
     random_dev_str <- cbind(
-        "", c("Intercept", x$effects$random), matrix2char(matrix(random_stddev, ncol = 1))
+        "", effects$random, matrix2char(matrix(random_stddev, ncol = 1))
     )
     random_dev_str[1, 1] <- effects$group
     random_dev_str <- rbind(
@@ -450,12 +529,13 @@ print.hgwrm <- function(x, decimal.fmt = "%.6f", ...) {
 #'
 #' @examples
 #' data(multisampling)
-#' model <- hgwr(formula = y ~ g1 + g2 + x1 + (z1 | group),
+#' model <- hgwr(formula = y ~ L(g1 + g2) + x1 + (z1 | group),
 #'               data = multisampling$data,
-#'               local.fixed = c("g1", "g2"),
 #'               coords = multisampling$coords,
 #'               bw = 10)
 #' summary(model)
+#' 
+#' @export 
 #'
 print.summary.hgwrm <- function(x, decimal.fmt = "%.6f", ...) {
     if (!inherits(x, "summary.hgwrm")) {
@@ -470,14 +550,50 @@ print.summary.hgwrm <- function(x, decimal.fmt = "%.6f", ...) {
     cat("   Data:", deparse(x$call[[3]]), fill = T)
     cat("\n")
 
+    ### Parameter Estimates
+    cat("Parameter estimates", fill = T)
+    cat("-------------------", fill = T)
+    cat("Fixed effects:", fill = T)
+    gfe <- x$effects$global.fixed
+    if (x$intercept$fixed) {
+        gfe <- c("Intercept", gfe)
+    }
+    pv_gfe <- x$significance$beta$pv
+    stars <- vapply(pv_gfe, pv2stars, rep(" ", n = length(pv_gfe)))
+    print.table.md(rbind(
+        c("", "Estimated", "Sd. Err", "t.val", "Pr(>|t|)", ""),
+        as.matrix(cbind(variable = gfe, x$significance$beta, stars = stars))
+    ), ...)
+    cat("\n")
+    cat("Local fixed effects:", fill = T)
+    lfe <- x$effects$local.fixed
+    if (x$intercept$local) {
+        lfe <- c("Intercept", lfe)
+    }
+    gamma_stats <- matrix2char(t(apply(x$gamma, 2, fivenum)))
+    gamma_stats_name <- c("Min", "1st Quartile", "Median", "3rd Quartile", "Max")
+    if (!is.null(x$significance$gamma)) {
+        pv_lfe <- x$significance$gamma$pv
+        gamma_stats <- cbind(
+            gamma_stats,
+            pv = matrix2char(pv_lfe),
+            stars = vapply(pv_lfe, pv2stars, rep(" ", n = length(pv_lfe)))
+        )
+        gamma_stats_name <- c(gamma_stats_name, "Pr(>|t|)", "")
+    }
+    gamma_str <- rbind(
+        c("", gamma_stats_name),
+        cbind(lfe, gamma_stats)
+    )
+    print.table.md(gamma_str, ...)
+    cat("\n")
+
     ### Diagnostics
     cat("Diagnostics", fill = T)
     cat("-----------", fill = T)
-    rsquared <- x$diagnostic$rsquared
-    diagnostic_mat <- matrix(c(rsquared), nrow = 1, ncol = 1)
-    diagnostic_chr <- rbind(
-        c("Rsquared"),
-        matrix2char(diagnostic_mat, decimal.fmt)
+    diagnostic_chr <- cbind(
+        names(x$diagnostic),
+        matrix2char(unlist(x$diagnostic), decimal.fmt)
     )
     print.table.md(diagnostic_chr, ...)
     cat("\n")
@@ -498,3 +614,30 @@ print.summary.hgwrm <- function(x, decimal.fmt = "%.6f", ...) {
     cat("Number of Obs:", nrow(x$frame), fill = T)
     cat("       Groups:", x$effects$group, ",", nrow(x$mu), fill = T)
 }
+
+#' Log likelihood function
+#' 
+#' @param object An `hgwrm` object.
+#' @param \dots Additional arguments.
+#' 
+#' @return An `logLik` instance used for S3 method `logLik()`.
+#' 
+#' @method logLik hgwrm
+#' @export 
+logLik.hgwrm <- function(object, ...) {
+    if (!inherits(object, "hgwrm")) {
+        stop("It's not an hgwrm object.")
+    }
+    n <- object$frame.parsed$y
+    p <- length(object$beta)
+    q <- ncol(object$mu)
+    enp_gwr <- 2 * object$trS[1] - object$trS[2]
+    enp_hlm <- p + q * (q + 1) / 2 + 1
+    enp <- enp_gwr + enp_hlm
+    val <- object$logLik
+    attr(val, "df") <- enp
+    attr(val, "nall") <- n
+    attr(val, "nobs") <- n
+    class(val) <- "logLik"
+    val
+}
